from http.client import HTTPSConnection, HTTPConnection
import json as JSON
from dsopz import util
from urllib.parse import urlparse, urlencode

class Error(Exception):
    """Exceptions"""

def req_text(method, url, params = None, headers = {}, expects = [200]):
    parsed = urlparse(url)
    host = parsed.netloc
    uri = parsed.path
    if isinstance(params, dict):
        params = urlencode(params) if headers.get('Content-Type', 'application/json').startswith('application/x-www-form-urlencoded') else JSON.dumps(params)
    if(parsed.query != ''):
        uri = uri + '?' + parsed.query
    conn = None
    if(parsed.scheme == 'https'):
        conn = HTTPSConnection(parsed.hostname, parsed.port or 443)
    else:
        conn = HTTPConnection(parsed.hostname, parsed.port or 80)
    if parsed.username != None:
        token = parsed.username + ':' + (parsed.password or '')
        token = 'Basic ' + base64.b64encode(token)
        headers['Authorization'] = token
    try:
        conn.request(method, uri, params, headers)
        response = conn.getresponse()
        if response.status not in expects:
            raise Error('Status: %d %s %sri' % (response.status, response.reason, response.read()))
        string = response.read()
        if string:
            string = string.decode('UTF-8')
        return {
            'status': response.status,
            'body': string
        }
    finally:
        util.close(conn)

def req_json(method, url, params = None, headers = {}, expects = [200]):
    resp = req_text(method, url, params, headers, expects)
    if not resp['body']:
        resp['body'] = None
    else:
        resp['body'] = JSON.loads(resp['body'])
    return resp

def __main():
    obj = req_json('GET', 'https://api.github.com/users/murer/keys', headers = {
        'User-Agent': 'dsopz'
    }, expects = [200])
    log.info(obj)

if __name__ == '__main__':
    __main()



from dsopz.config import config
from dsopz import io
from dsopz import dsutil
from dsopz.datastore import stream_entity, mutation
from dsopz.processor import blockify
from dsopz import util
from os import devnull
import logging as log

class Error(Exception):
    """Exceptions"""

def cmd_download():
    if (config.args.gql or config.args.query) and not config.args.dataset:
            raise Error('dataset is required for query or gql')
    if (config.args.resume or config.args.resume_gz) and (config.args.dataset or config.args.namespace):
        raise Error('dataset/namespace is not allowed for resume or resume_gz')
    header = dsutil.resolve_query(config.args.dataset, config.args.namespace, config.args.gql, config.args.query, config.args.resume, config.args.resume_gz)
    with io.jwriter(config.args.file, config.args.file_gz, append=config.args.append) as f:
        result = stream_entity(header['dataset'], header['namespace'], header['query'])
        query = next(result)
        if not config.args.append:
            f.write({'dataset': header['dataset'], 'namespace': header['namespace'], 'query': query})
        for line in result:
            f.write(line)

def cmd_kind():
    header = dsutil.resolve_query(config.args.dataset, config.args.namespace, 'select * from __kind__', None, None, None)
    with io.jwriter(config.args.file, config.args.file_gz, append=False) as f:
        result = stream_entity(header['dataset'], header['namespace'], header['query'])
        query = next(result)
        f.write({'dataset': header['dataset'], 'namespace': header['namespace'], 'query': query})
        for line in result:
            f.write(line)

def cmd_namespace():
    header = dsutil.resolve_query(config.args.dataset, None, 'select * from __namespace__', None, None, None)
    with io.jwriter(config.args.file, config.args.file_gz, append=False) as f:
        result = stream_entity(header['dataset'], header['namespace'], header['query'])
        query = next(result)
        f.write({'dataset': header['dataset'], 'namespace': header['namespace'], 'query': query})
        for line in result:
            f.write(line)

def cmd_upsert():
    skip = dsutil.resolve_mutation_skip(config.args.resume)
    log.info('Already processed: %s', skip)
    count = 0
    r = io.jwriter(config.args.resume, append=True) if config.args.resume else None
    try:
        with io.jreader(config.args.file, config.args.file_gz) as f:
            for block in blockify(f, 1000, lambda x: x.get('entity')):
                count = count + len(block)
                if count > skip:
                    log.info('Processing: %s', count)
                    mutation(config.args.dataset, config.args.namespace, upserts=block)
                    r.write(count)
    finally:
        util.close(r)

def cmd_rm():
    skip = dsutil.resolve_mutation_skip(config.args.resume)
    log.info('Already processed: %s', skip)
    count = 0
    r = io.jwriter(config.args.resume, append=True) if config.args.resume else None
    try:
        with io.jreader(config.args.file, config.args.file_gz) as f:
            for block in blockify(f, 1000, lambda x: x.get('entity', {}).get('key')):
                count = count + len(block)
                if count > skip:
                    log.info('Processing: %s', count)
                    mutation(config.args.dataset, config.args.namespace, removes=block)
                    r.write(count)
    finally:
        util.close(r)



subparser = config.add_parser('download', cmd_download)
subparser.add_argument('-d', '--dataset', help='dataset. Required for gql or query')
subparser.add_argument('-n', '--namespace', help='Namespace. Ignored for resume or resume-gz')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-g', '--gql', help='gql')
group.add_argument('-q', '--query', help='json query')
group.add_argument('-r', '--resume', help='json query and cursor from file')
group.add_argument('-rgz', '--resume-gz', help='json query and cursor from gzip file or - for stdin')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-f', '--file', help='output file or - for stdout')
group.add_argument('-fgz', '--file-gz', help='output gzip file or - for stdout')
subparser.add_argument('-a', '--append', action='store_true', help='append into file and do not write header')

subparser = config.add_parser('kind', cmd_kind)
subparser.add_argument('-d', '--dataset', required=True, help='dataset')
subparser.add_argument('-n', '--namespace', help='namespace')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-f', '--file', help='output file or - for stdout')
group.add_argument('-fgz', '--file-gz', help='output gzip file or - for stdout')

subparser = config.add_parser('namespace', cmd_namespace)
subparser.add_argument('-d', '--dataset', required=True, help='dataset')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-f', '--file', help='output file or - for stdout')
group.add_argument('-fgz', '--file-gz', help='output gzip file or - for stdout')

subparser = config.add_parser('upsert', cmd_upsert)
subparser.add_argument('-d', '--dataset', required=True, help='dataset')
subparser.add_argument('-n', '--namespace', help='namespace')
subparser.add_argument('-r', '--resume', help='file necessary to track and resume the process. Entities may be upsert twice anyway')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-f', '--file', help='input file or - for stdin')
group.add_argument('-fgz', '--file-gz', help='input gzip file or - for stdin')

subparser = config.add_parser('rm', cmd_rm)
subparser.add_argument('-d', '--dataset', required=True, help='dataset')
subparser.add_argument('-n', '--namespace', help='namespace')
subparser.add_argument('-r', '--resume', help='file necessary to track and resume the process. Entities may be upsert twice anyway')
group = subparser.add_mutually_exclusive_group(required=True)
group.add_argument('-f', '--file', help='input file or - for stdin')
group.add_argument('-fgz', '--file-gz', help='input gzip file or - for stdin')



from dsopz.http import req_json
from dsopz.oauth import oauth
from dsopz.config import config
import json as JSON

class Error(Exception):
    """Exceptions"""

def _set_partition(k, dataset, namespace):
    if not dataset:
        k.pop('partitionId')
        return
    k['partitionId'] = { 'projectId': dataset }
    if namespace:
        k['partitionId']['namespaceId'] = namespace

def ckey(k, dataset='dsopzproj', namespace=None):
    if len(k) % 2 != 0:
        raise Error('must be even %s' % (k))
    ret = {
        'partitionId': { 'projectId': dataset },
        'path': []
    }
    for i in range(0, len(k), 2):
        ret['path'].append({ 'kind': k[i], 'name': k[i+1] })
    if namespace:
        ret['partitionId']['namespaceId'] = namespace
    return ret

def cprop(name, type, value):
    return {
        name: {
            '%sValue' % (type): value
        }
    }

def centity(k, *props):
    entity = { 'key': k, 'properties': {}}
    for i in props:
        entity['properties'].update(i)
    return entity

def run_query(
        dataset,
        namespace,
        query):
    url = '%s/v1/projects/%s:runQuery' % (config.args.url, dataset)
    body = {
        'partitionId': {
            'projectId': dataset,
            'namespaceId': namespace
        }
    }
    if isinstance(query, dict):
        body['query'] = query
    else:
        body['gqlQuery'] = {
            'allowLiterals': True,
            'queryString': query
        }
    resp = req_json('POST', url, body, {
        'Authorization': 'Bearer %s' % (oauth.access_token())
    })
    ret = resp['body']
    ret['batch']['entityResults'] = ret['batch'].get('entityResults', [])
    ret['query'] = ret.get('query', query)
    for entity in ret['batch']['entityResults']:
        _set_partition(entity['entity']['key'], None, None)
    return ret

def lookup(dataset, keys):
    url = '%s/v1/projects/%s:lookup' % (config.args.url, dataset)
    body = { 'keys': keys }
    resp = req_json('POST', url, body, {
        'Authorization': 'Bearer %s' % (oauth.access_token())
    })
    ret = resp['body']
    if ret.get('deferred'):
        raise Error('idk what is deferred')
    ret = ret.get('found', [])
    for entity in ret:
        _set_partition(entity['entity']['key'], None, None)
    return ret

def commit(dataset, mutations):
    url = '%s/v1/projects/%s:commit' % (config.args.url, dataset)
    resp = req_json('POST', url, mutations, {
        'Authorization': 'Bearer %s' % (oauth.access_token())
    })
    ret = resp['body']
    return ret

def mutation(dataset, namespace, upserts=None, removes=None):
    body = { 'mode': 'NON_TRANSACTIONAL', 'mutations': [] }
    if upserts:
        for entity in upserts:
            _set_partition(entity['key'], dataset, namespace)
            body['mutations'].append({ 'upsert': entity })
    if removes:
        for k in removes:
            _set_partition(k, dataset, namespace)
            body['mutations'].append({ 'delete': k })
    return commit(dataset, body)

def stream_block(dataset, namespace, query):
    first = True
    while True:
        result = run_query(dataset, namespace, query)
        query = result['query']
        if first:
            yield query
            first = False
        if not result['batch']['entityResults']:
            return
        yield result
        query['startCursor'] = result['batch']['endCursor']

def stream_entity(dataset, namespace, query):
    result = stream_block(dataset, namespace, query)
    yield next(result)
    for k in result:
        for entity in k['batch']['entityResults']:
            yield entity



import threading
from concurrent.futures import Future

def _work(future, fn, args, kwargs):
    future.set_running_or_notify_cancel()
    try:
        ret = fn(*args, **kwargs)
    except BaseException as exc:
        future.set_exception(exc)
    else:
        future.set_result(ret)

def dispatch(fn, *args, **kwargs):
    ret = Future()
    thread = threading.Thread(target=_work, args=[ret, fn, args, kwargs])
    thread.start()
    return ret

def blockify(array, block_size, filter=None):
    ret = []
    for i in array:
        if filter:
            i = filter(i)
        if i != None:
            ret.append(i)
            if len(ret) >= block_size:
                yield ret
                ret = []
    if len(ret) > 0:
        yield ret



import sys
import time
import json as JSON
import os
from dsopz.config import config
from dsopz.http import req_json
from urllib.parse import urlencode, urlparse, parse_qs
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from dsopz import util
import webbrowser
import logging as log

class Error(Exception):
    """Exceptions"""

class OAuthHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        parsed = urlparse(self.path)
        _, port = self.server.socket.getsockname()
        params = parse_qs(parsed.query)
        if not params.get('code'):
            raise Error('Error %s', params)
        else:
            oauth._resume(port, params['code'][0])
        self.send_response(302)
        self.send_header('Location', 'http://github.com/murer/dsopz')
        self.send_header('Content-type','text/plain')
        self.end_headers()
        self.wfile.write('Ok'.encode('UTF-8'))

class OAuth(object):

    def _config(self):
        self._scopes = [
            'https://www.googleapis.com/auth/cloud-platform',
            'https://www.googleapis.com/auth/datastore',
            'https://www.googleapis.com/auth/userinfo.email'
        ]
        if hasattr(config.args, 'scopes') and config.args.scopes:
            self._scopes = config.args.scopes
        self._clientsecret = {
           "installed" : {
              "auth_uri" : "https://accounts.google.com/o/oauth2/auth",
              "client_secret" : "ZXEO84nRncBfhQnaEbMPWVfe",
              "auth_provider_x509_cert_url" : "https://www.googleapis.com/oauth2/v1/certs",
              "client_id" : "977896591232-sgpgg4ma9hbi7hvlgithb49ioj3o2sqj.apps.googleusercontent.com",
              "redirect_uris" : [
                 "urn:ietf:wg:oauth:2.0:oob",
                 "http://localhost"
              ],
              "project_id" : "dsopzproj",
              "token_uri" : "https://www.googleapis.com/oauth2/v3/token"
           }
        }
        if hasattr(config.args, 'client') and config.args.client:
            with open(config.args.client, 'r') as f:
                self._clientsecret = JSON.loads(f.read())
        self._auth_file = Path(os.path.expanduser(config.args.auth_file))

    def _prepare_url(self, redirect):
        return '%s?%s' % (self._clientsecret['installed']['auth_uri'],
            urlencode({
                'client_id': self._clientsecret['installed']['client_id'],
                'redirect_uri': redirect,
                'scope': ' '.join(self._scopes),
                'response_type': 'code',
                'approval_prompt': 'force',
                'access_type': 'offline'
            }))

    def _read_file(self):
        with open(str(self._auth_file), 'r') as f:
            return JSON.loads(f.read())

    def _write_file(self, content):
        util.makedirs(self._auth_file.parent)
        with open(str(self._auth_file), 'w') as f:
            f.write(JSON.dumps(content))
        log.info('Logged in', content)

    def _exchange_code(self, code, redirect):
        resp = req_json('POST', self._clientsecret['installed']['token_uri'], {
            'code': code,
            'client_id':  self._clientsecret['installed']['client_id'],
            'client_secret':  self._clientsecret['installed']['client_secret'],
            'redirect_uri': redirect,
            'grant_type': 'authorization_code'
        }, { 'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8' })
        now = now = int(time.time())
        expires_in = resp['body']['expires_in']
        resp['body']['created'] = now
        resp['body']['expires'] = now + expires_in
        return resp['body']

    def _fake(self):
        self._config()
        self._write_file({
            'expires' : int(time.time()) + 3600,
            'access_token' : 'dummytoken',
            'handler': 'fake'
        })

    def _login_text(self):
        url = self._prepare_url('urn:ietf:wg:oauth:2.0:oob')
        print(url)
        code = sys.stdin.readline().strip()
        resp = self._exchange_code(code, 'urn:ietf:wg:oauth:2.0:oob')
        resp['handler'] = 'installed'
        self._write_file(resp)

    def _login_browser(self):
        server = HTTPServer(('localhost', 0), OAuthHandler)
        _, port = server.socket.getsockname()
        url = self._prepare_url('http://localhost:%s/redirect_uri' % (port))
        try:
            webbrowser.open(url, new=1, autoraise=True)
            server.handle_request()
        finally:
            util.close(server.socket)

    def _resume(self, port, code):
        resp = self._exchange_code(code, 'http://localhost:%s/redirect_uri' % (port))
        resp['handler'] = 'browser'
        self._write_file(resp)

    def login(self):
        self._config()
        if config.args.text:
            self._login_text()
        else:
            self._login_browser()

    def _refresh_token(self, content):
        resp = req_json('POST', self._clientsecret['installed']['token_uri'], {
            'refresh_token': content['refresh_token'],
            'client_id': self._clientsecret['installed']['client_id'],
            'client_secret': self._clientsecret['installed']['client_secret'],
            'grant_type': 'refresh_token'
        }, { 'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8' })
        now = now = int(time.time())
        content['expires_in'] = resp['body']['expires_in']
        content['created'] = now
        content['expires'] = now + resp['body']['expires_in']
        content['access_token'] = resp['body']['access_token']
        content['id_token'] = resp['body']['id_token']
        self._write_file(content)


    def token(self):
        self._config()
        content = self._read_file()
        if not content:
            raise Error('You need to login')
        now = now = int(time.time())
        if now > content['expires'] - 60:
            self._refresh_token(content)
        return content

    def access_token(self):
        return self.token()['access_token']

    def print_token(self):
        content = self.token()
        if config.args.access_token:
            print(content['access_token'])
        else:
            print(JSON.dumps(content))

oauth = OAuth()

subparser = config.add_parser('login', oauth.login)
subparser.add_argument('-s', '--scopes', nargs='+', help='scopes')
subparser.add_argument('-t', '--text', action='store_true', help='text mode')
subparser.add_argument('-c', '--client', help='client secret json file')
subparser = config.add_parser('login-gce', oauth.login)
subparser = config.add_parser('login-serviceaccount', oauth.login)
subparser.add_argument('-f', '--file', required=True, help='service account json file')
subparser = config.add_parser('token', oauth.print_token)
subparser.add_argument('-a', '--access_token', action='store_true', help='complete json')



import argparse
import logging as log

log.basicConfig(
    format='%(asctime)s.%(msecs)03d [%(module)s:%(lineno)s] %(levelname)s %(message)s',
    datefmt='%m/%d/%Y %H:%M:%S',
    level=log.INFO)

class Config(object):

    def __init__(self):
        self.mods = {}
        self.parser = argparse.ArgumentParser(description='DSOpz')
        self.parser.add_argument('-a', '--auth-file', default='~/.dsopz/oauth.json', help='Local auth file')
        self.parser.add_argument('-u', '--url', default='https://datastore.googleapis.com/', help='URL')
        self.subparsers = self.parser.add_subparsers(dest="subparsers", title="Command")

    def add_parser(self, name, mod):
        self.mods[name] = mod
        return self.subparsers.add_parser(name)

    def parse_args(self, args=None):
        self.args = self.parser.parse_args(args)
        mod = self.mods[self.args.subparsers]
        mod()

config = Config()



from dsopz.config import config
from dsopz import io
from dsopz import util

def resolve_query(dataset, namespace, gql=True, query=None, plain=None, gz=None):
    if gql or query:
        ret = gql or query
        return { 'dataset': dataset, 'namespace': namespace, 'query': ret }
    if plain or gz:
        with io.jreader(plain, gz) as f:
            header = next(f)
            query = header['query']
            lastLine = None
            for line in f:
                lastLine = line
            query['startCursor'] = lastLine['cursor'] if lastLine else query['startCursor']
            return header
    raise Error('error')

def resolve_mutation_skip(resume):
    if not resume:
        return 0
    try:
        with io.jreader(resume) as r:
            ret = None
            for l in r:
                ret = l
            return ret
    except FileNotFoundError:
        return 0



import os
import errno
import json
import socket
import time
import socket

class Error(Exception):
	"""Exceptions"""

def close(obj):
    try:
        obj.close()
    except:
        """ Done """

def makedirs(directory):
    os.makedirs(str(directory), exist_ok=True)

def check_port(host, port):
    sock = None
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex((host, port))
        return result == 0
    finally:
        close(sock)

def wait_check_port(host, port, sleep=0.1, times=50):
    for i in range(times):
        if check_port(host, port):
            return
        time.sleep(sleep)
    raise Error('timeout %s:%s', host, port)



from dsopz import util
import gzip
import json as JSON
import sys

class Error(Exception):
    """Exceptions"""

class Writer(object):

    def __init__(self, f, gz=False, append=False):
        self._closable_plain = None
        self._closable_gz = None
        if not f:
            raise Error('not f')
        if f == '-':
            self._writer = sys.stdout.buffer
        elif isinstance(f, str):
            self._writer = open(f, 'ab' if append else 'wb')
            self._closable_plain = self._writer
        if gz:
            self._writer = gzip.open(self._writer, 'wb')
            self._closable_gz = self._writer

    def write(self, line):
        data = '%s\n' % (line)
        data = data.encode('UTF-8')
        self._writer.write(data)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def close(self):
        util.close(self._closable_gz)
        util.close(self._closable_plain)

class Reader(object):

    def __init__(self, f, gz=False):
        self._closable_plain = None
        self._closable_gz = None
        if not f:
            raise Error('not f')
        if f == '-':
            self._reader = sys.stdin.buffer
        elif isinstance(f, str):
            self._reader = open(f, 'rb')
            self._closable_plain = self._reader
        if gz:
            self._reader = gzip.open(self._reader, 'rb')
            self._closable_gz = self._reader

    def __next__(self):
        ret = next(self._reader)
        ret = ret.decode('UTF-8')
        ret = ret.strip()
        return ret

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def close(self):
        util.close(self._closable_gz)
        util.close(self._closable_plain)

class JWriter(Writer):

    def write(self, line):
        data = JSON.dumps(line)
        return super().write(data)

class JReader(Reader):

    def __next__(self):
        data = super().__next__()
        return JSON.loads(data)

def jreader(plain = None, gz = None):
    if not plain and not gz:
        raise Error('not plain and not gz')
    if plain and gz:
        raise Error('plain and gz')
    if gz:
        plain = gz
        gz = True
    else:
        gz = False
    return JReader(plain, gz)

def jwriter(plain = None, gz = None, append=False):
    if not plain and not gz:
        raise Error('not plain and not gz')
    if plain and gz:
        raise Error('plain and gz')
    if gz:
        plain = gz
        gz = True
    else:
        gz = False
    return JWriter(plain, gz=gz, append=append)



from distutils.core import setup

version = '2.0.beta1'

setup(
  name = 'dsopz',
  packages = ['dsopz','dsopz.cmd'],
  version = version,
  description = 'Google Datastore Operations',
  entry_points = {
      "console_scripts": [
          "dsopz = dsopz"
      ]
  },
  author = 'Paulo Henrique Murer',
  author_email = 'paulo.murer@gmail.com',
  url = 'https://github.com/murer/dsopz',
  download_url = 'https://github.com/murer/dsopz/tarball/dsopz-%s' % (version),
  keywords = ['gae', 'datastore', 'appengine'],
  classifiers = [],
)
